/*
:- module(infer_types,[get_type_from_imm/4, typed_pred/3,
		       get_types_from_exprs/3,
		       typify_query/2]).
*/
:- use_module(normalform).
:- multifile range_type/3.
:- multifile enum_type/2.
:- multifile (type_decl)/1.
:- dynamic range_type/3.
:- dynamic enum_type/2.
:- op(900,fx,type_decl).

%=range_type(Typname, Min, Max)
% Typname ist als Zahlenbereich von Min bis Max deklariert
range_type(jahr,1900,2050).
range_type(kalenderwoche,1,53).
%=enum_type(Typname,Konstruktoren)
% Typname ist mit ein Typ, dessen Elemente durch
% die genannten Funktoren realisiert sind.
enum_type(peano,[0,s(peano)]).
enum_type(janein,[ja,nein]).
enum_type(lst(A),[[A|lst(A)],[]]).

number_type(number).
number_type(RangeType):-range_type(RangeType,_,_).

more_special(RangeType,number):-range_type(RangeType,_,_).

%=error(Erklrung,Strafe) (*)
% falls noch mindestens <i>Strafe</i>
% Fehlerpunkte ber sind, werden diese abgezogen;
% wenn nicht, scheitert der Aufruf.
error(Reason,Penalty,
      ti_state(Vars,Errors,Pos,N),
      ti_state(Vars,[error(Pos,Reason)|Errors],Pos,N1)):-
	N1 is N-Penalty,
	N1>=0.

%=position(PosOld,PosNew) (*)
% holt und setzt die Positionsinformation
position(PosOld,PosNew,
	 ti_state(V,E,PosOld,N),
	 ti_state(V,E,PosNew,N)).

%=expect(DCG_Aufruf,Erklrung,Strafe) (*)
% falls <i>DCG_Aufruf</i> gelingt, so wird mit dessen Ergebnis
% fortgefahren, ansonsten wird dies als Fehler gewertet.
expect(What,Reason,Penalty,ListIn,ListOut):-
	call(What,ListIn,List1) *-> List1=ListOut ;
	error(Reason,Penalty,ListIn,ListOut).

%=expect1(Aufruf,Erklrung,Strafe) (*)
% falls <i>Aufruf</i> gelingt, so wird fortgefahren, ansonsten
% wird dies als Fehler gewertet.
expect1(What,Reason,Penalty,ListIn,ListOut):-
	call(What) *-> ListIn=ListOut ;
	error(Reason,Penalty,ListIn,ListOut).

%=get_type_from_imm(Term,Typ) (*)
% <i>Term</i> ist ein Term vom Typ <i>Typ</i>
get_type_from_imm('<var>'(VarName),TypeX,
		  ti_state(VarsIn,Errors,Pos,Penalty),
		  ti_state(VarsOut,Errors,Pos,Penalty)):-
	(   
	member(VarName=TypeX,VarsIn)->
	VarsIn=VarsOut ;
	VarsOut=[VarName=TypeX|VarsIn]).
get_type_from_imm('<a-var>'(_),_)-->[].

get_type_from_imm(X,RangeType)-->
	{ range_type(RangeType,Min,Max),
	  number(X),
	  X>=Min, X =< Max}.
get_type_from_imm(X,EnumType)-->
	{ once((enum_type(EnumType,Constructors),
		member(Cons,Constructors),
		X=..[Functor|Args],
		Cons=..[Functor|ArgTypes])) },
	match_typelist(Functor,Args,ArgTypes).
get_type_from_imm(X,number)-->{number(X)}.
get_type_from_imm(Term,NumberType)-->
	{ number_type(NumberType),
	  Term =.. [CompareOp,Var,Number],
	  compare_op(CompareOp),
	  Var='<var>'(_)
	},
	error(inline_compare(Term),5),
	get_type_from_imm(Var,NumberType),
	get_type_from_imm(Number,NumberType).

%=typify_arithmetic(Term) (*)
% <i>Term</i> ist ein arithmetischer Ausdruck,
% wie er bei =:=, is etc. vorkommt.
typify_arithmetic(X)-->
	get_type_from_imm(X,number).
typify_arithmetic(X)-->
	{
	 X=..[_|Args]
	},
	expect1(current_arithmetic_function(X),
		5,no_arithmetic_function(X)),
	typify_arithm_list(Args).
	 
typify_arithm_list([])-->[].
typify_arithm_list([A|Rest])-->
	typify_arithmetic(A),
	typify_arithm_list(Rest).
	  

%=compare_op(Op)
% Op ist ein Vergleichsoperator fr Zahlen
compare_op(<).
compare_op(>).
compare_op(=<).
compare_op(>=).

%=type_decl pred(Typ1,...,Typn)
% Die Argumente von <i>pred</i> haben die Typen
% Typ<i>i</i>
type_decl A = A.
type_decl member(A,lst(A)).
type_decl reverse(lst(A),lst(A)).
type_decl append(lst(A),lst(A),lst(A)).
type_decl sort(lst(A),lst(A)).

unify_types(Type1,Type2)-->{Type1=Type2}.
unify_types(Type1,Type2)-->
	{ once(A\=B; not(ground(A)); not(ground(B)))},
	error(should_be_same(Type1,Type2),5).

%=typed_pred(Term) (*)
% Term ist ein Aufruf, dessen Argumenten Typen zugewiesen werden.
typed_pred(Term)-->
	{ Term=..[Functor|Args],
	  type_decl Template,
	  Template=..[Functor|ArgTypes]},
	match_typelist(Functor,Args,ArgTypes).
typed_pred(Term)-->
	{ Term=..[Functor,Arg1,Arg2],
	  compare_op(Functor)},
	get_type_from_imm(Arg1,Type1),
	get_type_from_imm(Arg2,Type2),
	unify_types(Type1,Type2),
	(
	expect1(number_type(Type1),must_be_number(Type1),5),
	expect1(number_type(Type2),must_be_number(Type2),5)
	).
typed_pred(Term)-->
	{ Term=..[Functor1|Args],
	  type_decl Template,
	  Template=..[Functor|ArgTypes],
	  dwim_match(Functor1,Functor)},
	error(misspelled_predicate(Functor1,Functor),5),
	match_typelist(Functor,Args,ArgTypes).
typed_pred(Term)-->
	{ Term=..[Functor|_] },
	error(unknown_predicate(Functor),15).

%=match_typelist(Functor,Args,ArgTypes)
% sucht Typen fr Args
match_typelist(Functor,Args,ArgTypes)-->
	position(X,arg(0,X)),
	match_typelist(Functor,0,Args,ArgTypes),
	position(arg(_,X),X).
match_typelist(_,_,[],[])-->[].
match_typelist(Functor,N,[Arg|Args],[ArgType|ArgTypes])-->
	{succ(N,N1)},
	position(arg(N,X),arg(N1,X)),
	expect(get_type_from_imm(Arg,ArgType),type_expected(Arg,ArgType),5),
	match_typelist(Functor,N1,Args,ArgTypes).
match_typelist(Functor,N,Args,[_|ArgTypes])-->
	{succ(N,N1)},
	error(forgotten_arg(Functor,N),5),
	position(arg(N,X),arg(N1,X)),
	match_typelist(Functor,N1,Args,ArgTypes).
match_typelist(Functor,N,[Arg1,Arg2|Args],[ArgType1,ArgType2|ArgTypes])-->
	{succ(N,N1),succ(N1,N2)},
	error(swapped_args(Functor,N,N1),5),
	get_type_from_imm(Arg1,ArgType2),
	get_type_from_imm(Arg2,ArgType1),
	match_typelist(Functor,N2,Args,ArgTypes).

%=get_types_from_exprs(Exprs)
% Expr ist eine Liste von Termen aus einer
% Prolog-Anfrage bzw. einem Klauselkrper,
% auf denen Typinferenz betrieben wird.
get_types_from_exprs(Exprs)-->
	position(X,clause(0,X)),
	get_types_from_exprs1(Exprs),
	position(clause(_,X),X).
get_types_from_exprs1([])-->[].
get_types_from_exprs1([Expr|Rest])-->
	position(clause(N,X),clause(N1,X)),
	{succ(N,N1)},
	typed_pred(Expr),
	get_types_from_exprs1(Rest).

typify_query(Query,Errors):-
	typify_query(Query,0,40,Errors).
typify_query(Query,MinPenalty,_,Errors):-
	assert(trying(Query,MinPenalty)),
	bagof(Points-_Err,
	      VarsOut^PosOut^
	      get_types_from_exprs(Query,
				   ti_state([],[],query,MinPenalty),
				   ti_state(VarsOut,_Err,PosOut,Points)),
	      List),!,
	keysort(List,List1),
	assert(poss(MinPenalty,List)),
	last(List1,_-Errors).
typify_query(Query,MinPenalty,MaxPenalty,Errors):-
	MinPenalty<MaxPenalty,
	NewPenalty is MinPenalty+10,
	typify_query(Query,NewPenalty,MaxPenalty,Errors).
		     

