:- module(normalform,[consumeFile/2, consumeFiles/2,
		      consumeAtom1/2, consumeAtom/2,
		      eatFile/2,
		      replaceVars/1, normalize/2,
		      getPredicate/4,getPredicates/2,
		      putVariables/1]).

%=consumeFile(+Filename, -Liste)
% liest eine Datei mit Prolog-Klauseln
% und gibt die Klauseln in Liste zurck.
% Variablen werden dabei durch '<var>'(Variablenname) ersetzt.
consumeFile(Filename,Liste):-
	open(Filename,read,Fd),
	consumeStream(Fd,Liste,[]),
	close(Fd).

%=consumeFiles(+Filenames,-Liste)
% liest mehrere Dateien mit Prolog-Klauseln
% und verarbeitet sie dann wie consumeFile
consumeFiles([Filename|Rest],Liste):-
	open(Filename,read,Fd),
	consumeStream(Fd,Liste,LR),
	close(Fd),
	consumeFiles(Rest,LR).
consumeFiles([],[]).

%=consumeAtom(+Atom,-Klausel)
% liest einen Term aus einem Atom und wandelt die Variablen
% in Strukturen um.
consumeAtom1(Atom,Klausel):-
	catch(atom_to_term(Atom,Klausel,Bindings),Error,
	      (Klausel=erronous_term,
	       Bindings=[],
	       assert(user:current_warning(Error)))),
	replaceVars(Bindings),
	free_variables(Klausel,AnonymousVars),
	replaceAnonymous(AnonymousVars).

%=consumeAtom(-Atom,+Liste)
% Liest mehrere durch '.' getrennte Klauseln aus einem Atom
consumeAtom(Atom,Liste):-
	concat_atom(L1,'.',Atom),
	once(append(L2,[''],L1);
	     (append(L2,[Last],L1),
	      assert(user:current_warning(stuff_ignored(Last))))),
	maplist(consumeAtom1,L2,Liste).

consumeStream(Fd,XS,XR):-
	read_term(Fd,X,[variable_names(Vnames),variables(_)]),
	replaceVars(Vnames),
	(  X=end_of_file ->
		XS=XR ;
		XS=[X|Rest],consumeStream(Fd,Rest,XR)
	).

replaceVars([]).
replaceVars(['_'='<a-var>'(Name)|Rest]):-!,
	gensym('_',Name),
	replaceVars(Rest).
replaceVars([Name='<var>'(Name)|Rest]):-replaceVars(Rest).

replaceAnonymous([]).
replaceAnonymous(['<a-var>'(Name)|Rest]):-
	gensym('_',Name),
	replaceAnonymous(Rest).

%=putVariables(?Term)
% ersetzt freie Variablen in <i>Term</i> durch
% entsprechende Strukturen
putVariables(Term):-
	free_variables(Term,VarList),
	putVars(VarList,1).
putVars([],_).
putVars(['<var>'(VarName)|Rest],N):-
	atom_concat('Var',N,VarName),
	succ(N,N1),
	putVars(Rest,N1).


%=normalize(+Klausel,-NormalisierteKlausel)
% trennt Klauselkopf und Klauselkrper und
% wandelt den Klauselkrper in eine Liste um.
normalize((Head:-Body),Head-BodyList):- !,
	normalizeBody(Body,BodyList,[]).
normalize(Head,Head-[]).
normalizeBody((A,B))-->
	{!},
	normalizeBody(A),
	normalizeBody(B).
normalizeBody(X)-->[X].

%= getPredicate(Functor/Arity, ListIn, ListPred, ListNonPred)
% sucht die Klauseln zu einem Prdikat heraus
getPredicate(_,[],[],[]).
getPredicate(Functor/Arity,[Clause|Rest],[Clause|LPred],LNonPred):-
	Clause=Head-_,
	functor(Head,Functor,Arity),!,
	getPredicate(Functor/Arity,Rest,LPred,LNonPred).
getPredicate(Functor/Arity,[Clause|Rest],LPred,[Clause|LNonPred]):-
	getPredicate(Functor/Arity,Rest,LPred,LNonPred).


getPredicates([],[]).
getPredicates([Clause|Rest],[[Clause|LPred]|Others]):-
	Clause=Head-_,
	functor(Head,Functor,Arity),
	getPredicate(Functor/Arity,Rest,LPred,LNonPred),
	getPredicates(LNonPred,Others).


eatFile(FileName,PredList):-
	consumeFile(FileName,Unnormalized),
	maplist(normalize,Unnormalized,Normalized),
	getPredicates(Normalized,PredList).
